
// DUNGEON.CPP Version 1.0 -- Demonstrates tile scrolling through a dungeon
//		layout
// Written by Chris Lampton, 5/26/91
// All art by Mike Barrs
// All contents copyright (c) 1991 by the GAMERS Programming Workshop.

// A brief note on data structures:

//   This program is oriented around two arrays that contain two
// different forms of a 48 by 48 tile dungeon map. The first array,
// dungeon[], contains a description of the dungeon in terms of
// "quadtiles," each of which is constructed out of four of the
// small tiles contained in the file DUNTIL4B.GRF, arranged in a
// square. See the #defines below for a list of the 18 types of big
// tile. During initialization, this array is translated into a
// second array, dungeon2[], which contains a description of the
// dungeon in terms of the actual small tiles used on the screen.
// This system, which may or may not be used in the final version
// of REHACK, facilitates dungeon design while allowing more versa-
// tile use of the tile set.

#include <alloc.h>
#include <conio.h>
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>

#define  TILE_NUM 39                    // Number of small tiles
#define  TILE_HEIGHT 21                 // Height of small tile in pixels
#define  TILE_WIDTH 21                  // Width of small tile in pixels
#define  QUAD_TILE_NUM 18               // Number of quadtiles
#define  QUAD_TILE_WIDTH TILE_WIDTH*2   // Width of quadtile in pixels
#define  QUAD_TILE_HEIGHT TILE_HEIGHT*2 // Height of quadtile in pixels
#define  WIND_WIDTH 7                   // Width of dungeon window in small tiles
#define  WIND_HEIGHT 7                  // Height of dungeon window in small tiles
#define  DUNGEON_WIDTH 48               // Width of dungeon map in small tiles
#define  DUNGEON_HEIGHT 48              // Height of dungeon map in small tiles
#define  SCROLL_RATE 21                 // Number of pixels scrolled per frame

// The following #defines represent the 18 types of "quadtiles," each
// constructed out of four small tiles, that are the basic building blocks
// of the dungeon in this program.

#define  UL 0        // Upper left corner
#define  UR 1        // Upper right corner
#define  FL 2        // Floor
#define  LL 3        // Lower left corner
#define  LR 4        // Lower right corner
#define  HO 5        // Horizontal wall
#define  VE 6        // Vertical wall
#define  NT 7        // North 't' intersection
#define  ST 8        // South 't' intersection
#define  WT 9        // West 't' intersection
#define  ET 10       // East 't' intersection
#define  NE 11       // North wall end
#define  SE 12       // South wall end
#define  WE 13       // West wall end
#define  EE 14       // East wall end
#define  CC 15       // Central cross
#define  HD 16       // Horizontal door
#define  VD 17       // Vertical door

// Function prototypes:

void setmode(int);
void setpalette();
void translate();
void load_screen();
void blit(void far *,int,int,int,int,int,int,int,int);
void draw_window(int,int,int,int);
void draw_tile(int,int,int,int,int,int,int);
void putscreen();
void putwindow(int,int,int,int);
void wait_for_vbl(void);

unsigned char far *tile[TILE_NUM];      // Array of tile bitmaps
char far colregs[3*256];                // Array of color register values
unsigned char far screen_buffer[64000]; // Screen buffer
char quadtile[QUAD_TILE_NUM][4]={       // Array of quadtile definitions for translate()
	{0,10,3,4},
	{1,2,6,7},
	{5,5,5,5},
	{8,9,13,14},
	{11,12,15,16},
	{10,10,15,15},
	{3,12,3,12},
	{26,10,6,4},
	{17,9,15,15},
	{19,9,3,4},
	{24,12,6,12},
	{18,2,3,7},
	{25,12,13,27},
	{20,10,13,14},
	{22,2,15,23},
	{21,9,6,4},
	{10,10,29,30},
	{3,32,3,33}
};

// Dungeon map defined in terms of quadtiles (see above):

char dungeon[DUNGEON_HEIGHT/2][DUNGEON_WIDTH/2]={
	{UL,HO,HO,HO,HO,HO,HO,HO,HO,HO,HO,HO,HO,HO,HO,HO,HO,HO,HO,NT,HO,HO,HO,UR},
	{VD,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,VE,FL,FL,FL,VE},
	{VE,FL,UL,HO,HO,HO,HO,UR,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,WT,HO,UR,FL,VE},
	{VE,FL,VE,FL,FL,FL,FL,VE,FL,FL,UL,HO,HO,HO,HO,HO,HO,UR,FL,VE,FL,VE,FL,VE},
	{VE,FL,VD,FL,FL,FL,FL,VE,FL,FL,VE,FL,FL,FL,FL,FL,FL,VE,FL,VE,FL,SE,FL,VE},
	{VE,FL,VE,FL,FL,FL,FL,VE,FL,FL,VE,FL,FL,FL,FL,FL,FL,VE,FL,VE,FL,FL,FL,VE},
	{VE,FL,LL,HO,HO,HO,HO,LR,FL,FL,LL,HO,UR,FL,FL,UL,HO,LR,FL,VE,FL,NE,FL,VE},
	{VE,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,VE,FL,FL,VE,FL,FL,FL,VE,FL,VE,FL,VE},
	{VE,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,VE,FL,FL,VE,FL,FL,FL,WT,HO,LR,FL,VE},
	{VE,FL,UL,HO,HO,HO,UR,FL,FL,FL,FL,FL,VE,FL,FL,VE,FL,UL,HO,ET,FL,FL,FL,VE},
	{VE,FL,VE,FL,FL,FL,LL,HO,UR,FL,FL,FL,LL,HO,HD,LR,FL,VE,FL,VE,FL,FL,FL,VE},
	{VE,FL,VE,FL,FL,FL,FL,FL,VE,FL,FL,FL,FL,FL,FL,FL,FL,VE,FL,LL,UR,FL,FL,VE},
	{VE,FL,VE,FL,FL,FL,FL,FL,VE,FL,FL,FL,FL,UL,HO,HO,HO,ET,FL,FL,LL,EE,FL,VE},
	{VE,FL,LL,HO,HD,HO,HO,HO,LR,FL,FL,FL,FL,VE,FL,FL,FL,LL,HO,EE,FL,FL,FL,VE},
	{VE,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,VE,FL,NE,FL,FL,FL,FL,FL,FL,FL,VE},
	{VE,FL,UL,HO,HD,UR,FL,FL,FL,FL,FL,FL,FL,WT,HO,LR,FL,WE,HO,NT,HO,UR,FL,VE},
	{VE,FL,VE,FL,FL,LL,HO,UR,FL,FL,FL,FL,FL,VD,FL,FL,FL,FL,FL,VE,FL,VE,FL,VE},
	{VE,FL,VE,FL,FL,FL,FL,VE,FL,FL,FL,FL,FL,VE,FL,FL,FL,FL,FL,VE,FL,VE,FL,VE},
	{VE,FL,LL,HO,UR,FL,FL,LL,HO,UR,FL,FL,FL,WT,HO,HO,HO,UR,FL,VE,FL,VE,FL,VE},
	{VE,FL,FL,FL,VE,FL,FL,FL,FL,VD,FL,FL,FL,VE,FL,FL,FL,VE,FL,VE,FL,VE,FL,VE},
	{VE,FL,FL,FL,VE,FL,FL,FL,FL,VE,FL,FL,FL,VE,FL,NE,FL,VE,FL,VE,FL,VE,FL,VE},
	{VE,FL,FL,FL,LL,HO,HO,HO,HO,LR,FL,FL,FL,VE,FL,LL,HO,ST,HO,LR,FL,SE,FL,VE},
	{VE,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,FL,VE,FL,FL,FL,FL,FL,FL,FL,FL,FL,VE},
	{LL,HO,HO,HO,HO,HO,HO,HO,HO,HO,HO,HO,HO,ST,HO,HO,HO,HO,HO,HO,HO,HO,HO,LR},
};

char dungeon2[DUNGEON_HEIGHT][DUNGEON_WIDTH]; // Array for small tile map


int main()
{
	int	i,j,d,row=0,column=0,xoff,yoff;
	long  l;
	char far *temp;
	char	ch;
	FILE	*fp;

	setmode(0x13); // Set graphics mode to 13h

   // Open file containing bitmaps for tiles

	if ((fp=fopen("..\\pics\\duntil4b.grf","rb"))==NULL) {
		setmode(3);
		printf("Cannot open ..\\pics\\duntil4b.grf\n");
		exit(1);
	}

   // Load palette from file into colregs[]

	for (i=0; i<(3*256); i++) colregs[i]=getc(fp);

	load_screen(); // Load background screen into graphics buffer

   // Load tiles into tile[]

	for (i=0; i<TILE_NUM; i++) {
		tile[i]=(unsigned char far *)farmalloc(TILE_HEIGHT * TILE_WIDTH);
		for (j=0; j<(TILE_HEIGHT * TILE_WIDTH); j++)
			tile[i][j]=(unsigned char)getc(fp);
	}
	fclose(fp);
	translate();   // Translate quadtiles into small tiles
	setpalette();  // Set palette to palette in tile file
	draw_window(0,0,0,0); // Draw initial contents of dungeon window in graphics buffer
	putscreen();   // Move graphics buffer to the screen

   // Main program loop
   // Waits for cursor arrow to be pressed, then scrolls dungeon window

	do {
		ch=getch();
		if (ch==0) {
			ch=getch();
			switch (ch) {
				case 72: { // Up arrow, so scroll up
					if (row>0) {
						--row;
						for (yoff=TILE_HEIGHT-SCROLL_RATE; yoff>0; yoff-=SCROLL_RATE)
							draw_window(column,row,0,yoff);
							putwindow(10,8,WIND_WIDTH*TILE_WIDTH,WIND_HEIGHT*TILE_HEIGHT);
					}
					break;
				}
				case 80: {	// Down arrow, so scroll down
					if (row<(DUNGEON_HEIGHT-WIND_HEIGHT)) {
						for (yoff=SCROLL_RATE; yoff<TILE_HEIGHT; yoff+=SCROLL_RATE)
							draw_window(column,row,0,yoff);
							putwindow(10,8,WIND_WIDTH*TILE_WIDTH,WIND_HEIGHT*TILE_HEIGHT);
						row++;
					}
					break;
				}
				case 75: {	// Left arrow, so scroll left
					if (column>0) {
						--column;
						for (xoff=TILE_WIDTH-SCROLL_RATE; xoff>0; xoff-=SCROLL_RATE)
							draw_window(column,row,xoff,0);
							putwindow(10,8,WIND_WIDTH*TILE_WIDTH,WIND_HEIGHT*TILE_HEIGHT);
					}
					break;
				}
				case 77: { // Right arrow, so scroll right
					if (column<(DUNGEON_WIDTH-WIND_WIDTH)) {
						for (xoff=SCROLL_RATE; xoff<TILE_WIDTH; xoff+=SCROLL_RATE)
							draw_window(column,row,xoff,0);
							putwindow(10,8,WIND_WIDTH*TILE_WIDTH,WIND_HEIGHT*TILE_HEIGHT);
						column++;
					}
					break;
				}
			}
			draw_window(column,row,0,0);
			putwindow(10,8,WIND_WIDTH*TILE_WIDTH,WIND_HEIGHT*TILE_HEIGHT);
		}
	} while (ch!=27);  // Terminate if ESC pressed
	setmode(3);        // Restore DOS screen mode

	for (i=0; i<TILE_NUM; i++) {
		farfree( tile[ i ] );
	}

	return 0;
}

// translate()

// Translates dungeon array from quadtile format in dungeon[] to
// small tile format in dungeon2[].

void translate()
{
	int	x,y;

   // Translate quadtiles to small tiles

	for (x=0;x<DUNGEON_WIDTH/2;x++)
		for (y=0;y<DUNGEON_HEIGHT/2;y++) {
			dungeon2[y*2][x*2]=quadtile[dungeon[y][x]][0];
			dungeon2[y*2][x*2+1]=quadtile[dungeon[y][x]][1];
			dungeon2[y*2+1][x*2]=quadtile[dungeon[y][x]][2];
			dungeon2[y*2+1][x*2+1]=quadtile[dungeon[y][x]][3];
		}

   // Add corners of doors to tiles next to door tiles

	for (x=0;x<DUNGEON_WIDTH/2;x++)
		for (y=0;y<DUNGEON_HEIGHT/2;y++) {
			if (dungeon[y][x]==HD) dungeon2[y*2+1][x*2-1]=28;
			if (dungeon[y][x]==VD) dungeon2[y*2+2][x*2+1]=34;
		}
}

// load_screen()

// Loads background screen from disk into screen_buffer[]

void load_screen()
{
	long	i;
	unsigned char far *screen;
	FILE *fp;

	if ((fp=fopen("..\\pics\\mbs08-b.grf","rb"))==NULL) {
		setmode(3);
		printf("Cannot open ..\\pics\\mbs08-b.grf\n");
		exit(1);
	}
	for (i=0; i<(3*256); i++) getc(fp);
	for (i=0; i<64000; i++) screen_buffer[i]=(unsigned char)getc(fp);
	fclose(fp);
}

// draw_window(int xtile,int ytile,int xoff,int yoff)

// Draws dungeon window with dimensions (in small tiles) of WIND_WIDTH,
// WIND_HEIGHT at screen coordinates xpos, ypos. The first row and
// column of tiles is offset by xoff, yoff pixels.

void draw_window(int xtile,int ytile,int xoff,int yoff)
{
	int	x,y,xpos,ypos,xoff1,yoff1,xext,yext;

	for (x=0; x<(WIND_WIDTH+1);x++)
		for (y=0; y<(WIND_HEIGHT+1);y++) {
			xpos=(x==0)?10:10+x*TILE_WIDTH-xoff;
			ypos=(y==0)?8:8+y*TILE_HEIGHT-yoff;
			if (x==0) {
				xoff1=xoff;
				xext=TILE_WIDTH-xoff;
			}
			else {
				xoff1=0;
				xext=TILE_WIDTH;
			};
			if (y==0) {
				yoff1=yoff;
				yext=TILE_HEIGHT-yoff;
			}
			else {
				yoff1=0;
				yext=TILE_HEIGHT;
			};
			if (x==WIND_WIDTH) xext=xoff;
			if (y==WIND_HEIGHT) yext=yoff;
			if	(!(((x==WIND_WIDTH)&&(xoff==0))||((y==WIND_HEIGHT)&&(yoff==0))))
				draw_tile((int)dungeon2[ytile+y][xtile+x],xpos,ypos,xoff1,yoff1,xext,yext);
		}
}

// draw_tile(int tnum,int xpos,int ypos,int xoff,int yoff,int xext,int yext)

// Draw xext,yext pixels of tile tnum starting at pixel xoff,yoff at
// screen coordinates xpos,ypos

void draw_tile(int tnum,int xpos,int ypos,int xoff,int yoff,int xext,int yext)
{
	blit(tile[tnum],xpos,ypos,TILE_WIDTH,TILE_HEIGHT,xoff,yoff,xext,yext);
}

// blit(void far *tile,int xpos,int ypos,int xsize,int ysize,int xoff,int yoff,int xext,int yext)

// Draw rectangular bitmap, or portion of bitmap, in screen_buffer[]
// Parameters:
//   *tile - pointer to bitmap
//   xpos  - x coordinate to draw bitmap at
//   ypos  - y coordinate to draw bitmap at
//   xsize - horizontal dimension, in pixels, of bitmap
//   ysize - vertical dimension, in pixels, of bitmap
//   xoff  - horizontal offset within bitmap, measured in pixels, to start drawing at
//   yoff  - vertical offset within bitmap, measured in pixels, to start drawing at
//   xext  - number of pixels to draw horizontally in bitmap
//   yext  - number of pixels to draw vertically in bitmap

void blit(void far *tile,int xpos,int ypos,int xsize,int ysize,int xoff,int yoff,int xext,int yext)
{
	int	screen;
	int   offscr,segscr,offtil,segtil,xjump;

	offscr=FP_OFF(screen_buffer);
	segscr=FP_SEG(screen_buffer);
	offtil=FP_OFF(tile)+xoff+yoff*xsize;
	segtil=FP_SEG(tile);
	screen=offscr+320*ypos+xpos;
	xjump=xsize-xext;
	asm {
		mov	dx,es          // Save segment registers
		push	dx
		mov	dx,ds
		push	dx
		mov	dx,segscr		// Get screen buffer segment in ES
		mov	es,dx
		mov	dx,segtil		// Get tile segment in DS
		mov	ds,dx
		mov	si,offtil      // Point SI at tile bitmap plus xoff,yoff
		mov	di,screen      // Point DI at address in screen buffer
		mov	dx,yext        // Get line count in DX
		cld
	}
	loop1:
	asm {
		mov	cx,xext        // Get pixel count in CX
		push  di             // Save screen buffer address
		rep	movsb          // Move one line of tile to screen buffer
		pop	di             // Restore screen buffer address
		add	di,320         // ...and advance it to next line on screen
		add	si,xjump       // Advance SI to next line of tile bitmap
		dec	dx             // Count off one line
		jnz	loop1          // If more lines in tile, loop back and draw them
		pop	dx             // Else restore segment registers and return
		mov	ds,dx
		pop	dx
		mov   es,dx
	}
}

// putwindow(int xpos,int ypos,int xsize,int ysize)

// Move rectangle at xpos,ypos with dimensions xsize,ysize from
// screen_buffer[] to video memory

void putwindow(int xpos,int ypos,int xsize,int ysize)
{
	int   offbuf,segbuf,offscr,segscr,xjump;

	wait_for_vbl();
	offbuf=FP_OFF(screen_buffer);
	segbuf=FP_SEG(screen_buffer);
	offscr=320*ypos+xpos;
	offbuf=offbuf+offscr;
	asm {
		mov	dx,es          // Save segment registers
		push	dx
		mov	dx,ds
		push	dx
		mov	dx,0a000h		// Get screen segment in ES
		mov	es,dx
		mov	dx,segbuf		// Get screen buffer segment in DS
		mov	ds,dx
		mov	si,offbuf      // Point SI at window address within buffer
		mov	di,offscr      // Point DI at window address on screen
		mov	dx,ysize       // Get line count into DX
		cld }
	loop1:
	asm {
		mov	cx,xsize       // Get pixel width of window into CX
		push  di             // Save screen and buffer addresses
		push	si
		rep	movsb          // Move one line of window to screen
		pop	si             // Restore screen and buffer addresses
		pop	di
		add	si,320         // ...and advance them to next line
		add	di,320
		dec	dx             // Count off one line
		jnz	loop1          // If more lines in window, loop back and draw them
		pop	dx             // Else restore segment registers and return
		mov	ds,dx
		pop	dx
		mov   es,dx
	}
}


void putscreen()
{
	int offbuf,segbuf;

	offbuf=FP_OFF(screen_buffer);
	segbuf=FP_SEG(screen_buffer);
	asm {
		mov	dx,es          // Save segment registers
		push	dx
		mov	dx,ds
		push	dx
		mov	dx,0a000h      // Get screen segment into ES
		mov	es,dx
     	mov	dx,segbuf      // Get buffer segment into DS
		mov	ds,dx
		mov	si,offbuf      // Point SI at start of buffer
		mov	di,0           // Point DI at start of screen
		mov	cx,32000       // Count 32,000 words
		rep	movsw          // Move 32,000 words from buffer to screen
		pop	dx             // Restore segment registers
		mov	ds,dx
		pop	dx
		mov	es,dx
	}
}

// setmode(int mode)

// Calls video interrupt to set graphics adaptor to mode

void setmode(int mode)
{
	union REGS inregs, outregs;
	struct SREGS segregs;

	inregs.h.ah = 0x00;
	inregs.h.al = mode;
	int86x(0x10, &inregs, &outregs, &segregs);
}

// setpalette()

// Sets all 256 VGA color registers to RGB values in colregs[]

void setpalette()
{
	int	offcol,segcol;

	offcol=FP_OFF(colregs);
	segcol=FP_SEG(colregs);
	asm{
		mov	dx,es
		push	dx
		mov	dx,segcol
		mov	es,dx
		mov	ah,10h
		mov	al,12h
		mov	bx,0
		mov	cx,100h
		mov	dx,offcol
		int	10h
		pop	dx
		mov	es,dx
	}
}

// wait_for_vbl(void)

// Loop to delay until next vertical blank

void wait_for_vbl(void)
{
	while ((inport(0x3da)&8)==0);
}
